var _a;
import { MAX_RECOMMENDED_TX_SIZE, cojsonInternals } from "cojson";
import { ItemsSym, Account, CoValueBase, Ref, inspect, co, SchemaInit, isRefEncoded, loadCoValue, subscribeToCoValue, ensureCoValueLoaded, subscribeToExistingCoValue, } from "../internal.js";
/** @category CoValues */
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export class CoStream extends CoValueBase {
    static Of(item) {
        var _b, _c;
        return _c = class CoStreamOf extends _a {
                constructor() {
                    super(...arguments);
                    this[_b] = item;
                }
            },
            _b = co.items,
            _c;
    }
    get _schema() {
        return this.constructor._schema;
    }
    get byMe() {
        return this[this._loadedAs.id];
    }
    get inCurrentSession() {
        return this.perSession[this._loadedAs.sessionID];
    }
    constructor(options) {
        super();
        if (options && "fromRaw" in options) {
            Object.defineProperties(this, {
                id: {
                    value: options.fromRaw.id,
                    enumerable: false,
                },
                _raw: { value: options.fromRaw, enumerable: false },
            });
        }
        return new Proxy(this, CoStreamProxyHandler);
    }
    static create(init, options) {
        const instance = new this({ init, owner: options.owner });
        const raw = options.owner._raw.createStream();
        Object.defineProperties(instance, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _raw: { value: raw, enumerable: false },
        });
        if (init) {
            instance.push(...init);
        }
        return instance;
    }
    push(...items) {
        for (const item of items) {
            this.pushItem(item);
        }
    }
    pushItem(item) {
        const itemDescriptor = this._schema[ItemsSym];
        if (itemDescriptor === "json") {
            this._raw.push(item);
        }
        else if ("encoded" in itemDescriptor) {
            this._raw.push(itemDescriptor.encoded.encode(item));
        }
        else if (isRefEncoded(itemDescriptor)) {
            this._raw.push(item.id);
        }
    }
    toJSON() {
        const itemDescriptor = this._schema[ItemsSym];
        const mapper = itemDescriptor === "json"
            ? (v) => v
            : "encoded" in itemDescriptor
                ? itemDescriptor.encoded.encode
                : (v) => v && v.id;
        return {
            id: this.id,
            _type: this._type,
            ...Object.fromEntries(Object.entries(this).map(([account, entry]) => [
                account,
                mapper(entry.value),
            ])),
            in: Object.fromEntries(Object.entries(this.perSession).map(([session, entry]) => [
                session,
                mapper(entry.value),
            ])),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    static schema(def) {
        this._schema || (this._schema = {});
        Object.assign(this._schema, def);
    }
    /** @category Subscription & Loading */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /** @category Subscription & Loading */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /** @category Subscription & Loading */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /** @category Subscription & Loading */
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
}
_a = CoStream;
(() => {
    _a.prototype._type = "CoStream";
})();
function entryFromRawEntry(accessFrom, rawEntry, loadedAs, accountID, itemField) {
    return {
        get value() {
            if (itemField === "json") {
                return rawEntry.value;
            }
            else if ("encoded" in itemField) {
                return itemField.encoded.decode(rawEntry.value);
            }
            else if (isRefEncoded(itemField)) {
                return this.ref?.accessFrom(accessFrom, rawEntry.by +
                    rawEntry.tx.sessionID +
                    rawEntry.tx.txIndex +
                    ".value");
            }
            else {
                throw new Error("Invalid item field schema");
            }
        },
        get ref() {
            if (itemField !== "json" && isRefEncoded(itemField)) {
                const rawId = rawEntry.value;
                return new Ref(rawId, loadedAs, itemField);
            }
            else {
                return undefined;
            }
        },
        get by() {
            return (accountID &&
                new Ref(accountID, loadedAs, { ref: Account, optional: false })?.accessFrom(accessFrom, rawEntry.by +
                    rawEntry.tx.sessionID +
                    rawEntry.tx.txIndex +
                    ".by"));
        },
        madeAt: rawEntry.at,
        tx: rawEntry.tx,
    };
}
export const CoStreamProxyHandler = {
    get(target, key, receiver) {
        if (typeof key === "string" && key.startsWith("co_")) {
            const rawEntry = target._raw.lastItemBy(key);
            if (!rawEntry)
                return;
            const entry = entryFromRawEntry(receiver, rawEntry, target._loadedAs, key, target._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = target._raw.itemsBy(key);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(receiver, rawEntry.value, target._loadedAs, key, target._schema[ItemsSym]);
                        }
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    })();
                },
            });
            return entry;
        }
        else if (key === "perSession") {
            return new Proxy({}, CoStreamPerSessionProxyHandler(target, receiver));
        }
        else {
            return Reflect.get(target, key, receiver);
        }
    },
    set(target, key, value, receiver) {
        var _b;
        if (key === ItemsSym &&
            typeof value === "object" &&
            SchemaInit in value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                value[SchemaInit];
            return true;
        }
        else {
            return Reflect.set(target, key, value, receiver);
        }
    },
    defineProperty(target, key, descriptor) {
        var _b;
        if (descriptor.value &&
            key === ItemsSym &&
            typeof descriptor.value === "object" &&
            SchemaInit in descriptor.value) {
            (_b = target.constructor)._schema || (_b._schema = {});
            target.constructor._schema[ItemsSym] =
                descriptor.value[SchemaInit];
            return true;
        }
        else {
            return Reflect.defineProperty(target, key, descriptor);
        }
    },
    ownKeys(target) {
        const keys = Reflect.ownKeys(target);
        for (const accountID of target._raw.accounts()) {
            keys.push(accountID);
        }
        return keys;
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
};
const CoStreamPerSessionProxyHandler = (innerTarget, accessFrom) => ({
    get(_target, key, receiver) {
        if (typeof key === "string" && key.includes("session")) {
            const sessionID = key;
            const rawEntry = innerTarget._raw.lastItemIn(sessionID);
            if (!rawEntry)
                return;
            const by = cojsonInternals.accountOrAgentIDfromSessionID(sessionID);
            const entry = entryFromRawEntry(accessFrom, rawEntry, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                ? by
                : undefined, innerTarget._schema[ItemsSym]);
            Object.defineProperty(entry, "all", {
                get: () => {
                    const allRawEntries = innerTarget._raw.itemsIn(sessionID);
                    return (function* () {
                        while (true) {
                            const rawEntry = allRawEntries.next();
                            if (rawEntry.done)
                                return;
                            yield entryFromRawEntry(accessFrom, rawEntry.value, innerTarget._loadedAs, cojsonInternals.isAccountID(by)
                                ? by
                                : undefined, innerTarget._schema[ItemsSym]);
                        }
                        // eslint-disable-next-line @typescript-eslint/no-explicit-any
                    })();
                },
            });
            return entry;
        }
        else {
            return Reflect.get(innerTarget, key, receiver);
        }
    },
    ownKeys() {
        return innerTarget._raw.sessions();
    },
    getOwnPropertyDescriptor(target, key) {
        if (typeof key === "string" && key.startsWith("co_")) {
            return {
                configurable: true,
                enumerable: true,
                writable: false,
            };
        }
        else {
            return Reflect.getOwnPropertyDescriptor(target, key);
        }
    },
});
/** @category CoValues */
export class BinaryCoStream extends CoValueBase {
    constructor(options) {
        super();
        let raw;
        if ("fromRaw" in options) {
            raw = options.fromRaw;
        }
        else {
            const rawOwner = options.owner._raw;
            raw = rawOwner.createBinaryStream();
        }
        Object.defineProperties(this, {
            id: {
                value: raw.id,
                enumerable: false,
            },
            _type: { value: "BinaryCoStream", enumerable: false },
            _raw: { value: raw, enumerable: false },
        });
    }
    static create(options) {
        return new this(options);
    }
    getChunks(options) {
        return this._raw.getBinaryChunks(options?.allowUnfinished);
    }
    start(options) {
        this._raw.startBinaryStream(options);
    }
    push(data) {
        this._raw.pushBinaryStreamChunk(data);
    }
    end() {
        this._raw.endBinaryStream();
    }
    toBlob(options) {
        const chunks = this.getChunks({
            allowUnfinished: options?.allowUnfinished,
        });
        if (!chunks) {
            return undefined;
        }
        return new Blob(chunks.chunks, { type: chunks.mimeType });
    }
    static async loadAsBlob(id, as, options) {
        const stream = await this.load(id, as, []);
        return stream?.toBlob({
            allowUnfinished: options?.allowUnfinished,
        });
    }
    static async createFromBlob(blob, options) {
        const stream = this.create({ owner: options.owner });
        const start = Date.now();
        const data = new Uint8Array(await blob.arrayBuffer());
        stream.start({
            mimeType: blob.type,
            totalSizeBytes: blob.size,
            fileName: blob instanceof File ? blob.name : undefined,
        });
        const chunkSize = MAX_RECOMMENDED_TX_SIZE;
        let lastProgressUpdate = Date.now();
        for (let idx = 0; idx < data.length; idx += chunkSize) {
            stream.push(data.slice(idx, idx + chunkSize));
            if (Date.now() - lastProgressUpdate > 100) {
                options.onProgress?.(idx / data.length);
                lastProgressUpdate = Date.now();
            }
            await new Promise((resolve) => setTimeout(resolve, 0));
        }
        stream.end();
        const end = Date.now();
        console.debug("Finished creating binary stream in", (end - start) / 1000, "s - Throughput in MB/s", (1000 * (blob.size / (end - start))) / (1024 * 1024));
        options.onProgress?.(1);
        return stream;
    }
    toJSON() {
        return {
            id: this.id,
            _type: this._type,
            ...this.getChunks(),
        };
    }
    [inspect]() {
        return this.toJSON();
    }
    /** @category Subscription & Loading */
    static load(id, as, depth) {
        return loadCoValue(this, id, as, depth);
    }
    /** @category Subscription & Loading */
    static subscribe(id, as, depth, listener) {
        return subscribeToCoValue(this, id, as, depth, listener);
    }
    /** @category Subscription & Loading */
    ensureLoaded(depth) {
        return ensureCoValueLoaded(this, depth);
    }
    /** @category Subscription & Loading */
    subscribe(depth, listener) {
        return subscribeToExistingCoValue(this, depth, listener);
    }
}
//# sourceMappingURL=coStream.js.map